
function [samples, Hbar, xbar] = GASS(N, K, test_func, M, Hl, alpha, rho)
% usage example:
%   [S,H,X]=GASS(1000,500,@flip_sphere,50,-1e10, 1, 0.05);
% 
% N: # of samples
% K: # of iterations
% test_func: @+function name of the test, a function to be maximized
% M: dimension of the test function
% Hl: lower bound of test_func, meaning that test_func-Hl >0
% alpha: step size parameter
% rho: quantile parameter
% samples: number of samples
% Hbar: optimal value
% xbar: candidate solution


% Initialization
mu_k = rand(1,M)*60-30;  % -30 < mu0 < 30
var_k = 1000*ones(1,M);   % the variance vector
theta1_k=mu_k./var_k;
theta2_k=-0.5*ones(1,M)./var_k;
theta_k=[theta1_k, theta2_k];

X_k=randn(N,M)*diag(sqrt(var_k))+ones(N,1)*mu_k;
stop=0;
k=1;
samples=zeros(K,1);
Hbar = zeros(K,1);
xbar = zeros(K,M);
Ntotal=0;

while (stop==0)&&(k<=K) 
    gamma_k=alpha/k^0.05;  % stepsize
    Ntotal=Ntotal+N;
    samples(k) = Ntotal;


    G = feval(test_func,X_k)-Hl;
    if G<0 % this is a test for whether the user has correctly specified the parameter Hl
        error('Hl is too big, and cannot gaurantee: test_func - Hl >0');
    end
    
    
    % approximate the solution
    [Hbar(k), indx] = max(G);
    xbar(k,:) = X_k(indx,:);

    if isnan(Hbar(k))==1  
        Hbar(k)=Hbar(k-1);
        xbar(k,:) =xbar(k-1,:);
        %fprintf('nan \n');
        break
    end

    if k>2
    if Hbar(k-1)>max(G)     
        Hbar(k)=Hbar(k-1);
        xbar(k,:) =xbar(k-1,:);
    end 
    end
    
    %--shape function---take the elite samples---
    [G_sort, I_sort] = sort(G,'descend');   % sorted performances
    gm = G_sort(ceil(N*rho));
    G_u=(G>gm).*G;
    %G_u=G./(1+exp(-1e5*(G-gm)));
    
    % esitmate gradient and hessian
    w_k=G_u/sum(G_u);
    CX_k=[X_k,X_k.*X_k];
    grad_k=w_k'*CX_k-[mu_k, var_k + mu_k.*mu_k];  % use the exact expectation mu_k instead of the estimate mean(CX_k)
    Hes_k=-cov(CX_k);
    %Hes_inv_k=inv(Hes_k + 1e-30.*eye(2*M));  % ensure positive-definiteness
    Hes_inv_k=(Hes_k + 1e-30.*eye(2*M))\diag(ones(1,2*M)); 
    if norm(grad_k) < 1e-30
        fprintf('grad small! \n');
        stop=1; break;
    else
    % update the parameter using an SA iteration
    theta_k=theta_k-gamma_k*(Hes_inv_k*grad_k')';
    theta1_k=theta_k(1:M);
    theta2_k=theta_k(M+1:2*M);
    var_k=-0.5./theta2_k;
    mu_k=theta1_k.*var_k;
    % generate new samples
    X_k=randn(N,M)*diag(sqrt(abs(var_k)))+ones(N,1)*mu_k;
    end  %end if

    k=k+1;
    clear G G_u w_k theta1_k theta2_k CX_k grad_k Hes_k Hes_inv_k;
end %end while

for l=k:K
samples(l) = Ntotal;
Hbar(l)=Hbar(k-1);
xbar(l,:) =xbar(k-1,:);
Ntotal=Ntotal+N;
end
Hbar=Hbar+Hl;